////////////////////////////////////////////////////
// Jet Header File.
// Consists of the C++ bindings (class definitions, 
// macros and functions) for the jet language.
// Not to be used for multiple file compilation.
// Anthony Savidis, Yiannis Lilis 2005.
////////////////////////////////////////////////////
// Revised June 2006, A. Savidis, changed syntax patterns to better emulate
// assembly language style.
// Project: JET Suite (Execution visualizer, Savidis, 2005).
//

#ifndef JET_H
#define JET_H

#pragma warning(disable:4786)

#include <string>
#include <vector>
#include <functional>
#include <algorithm>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>

#if	defined(HAS_VISUALIZER) && (defined(CONSOLE_VISUALIZER) || defined(GRAPHICAL_VISUALIZER))
#include "VisualiserSuper.h"
#include "Specs.h"
#else

// If we do not have a visualizer we simply strip-off the
// notification calls.
//
#define NOTIFY_REGASSIGN(r, c)
#define NOTIFY_MEMASSIGN(m, c)
#define NOTIFY_RFORMAT(n, ra, rb, rc, l)
#define NOTIFY_IFORMAT(n, ra, rb, im, l)
#define NOTIFY_LFORMAT(n, ra, im, l)
#define NOTIFY_BRANCH(n, id, ra, rb, la, l)
#define NOTIFY_JUMP(n, la, l)

#define	NOTIFY_PROGRAMSTART() \
		OUTPUT_FUNCTION("JET starts running the program.\n")

#define	NOTIFY_PROGRAMEND()	\
	OUTPUT_FUNCTION("JET finished the program normally.\n");

#define	NOTIFY_PROGRAMERROR(msg) \
	OUTPUT_FUNCTION("Error: %s\nJET has interrupted program execution.\n", msg);

// Default std I/O functions and clean-up.
//
#define	INPUT_FUNCTION	fgets
#define	OUTPUT_FUNCTION	printf
inline	void CLEANUP_FUNCTION (void) { exit(1); }

// No need for the spec file when no visualizer is declared.
//
#define MEMORY_SIZE			1024
#define TOTAL_REGISTERS		32
#define BITS_PER_REGISTER	32

#endif	// HAS_VISUALIZER

void ReadInputString (char* buff, unsigned len)  {
	do
		;
	while (!INPUT_FUNCTION(buff, len-1, stdin) || !buff[0] || buff[0]=='\n');
	
	// Strip-off trailing new line character.
	//
	unsigned i = strlen(buff);	
	if (buff[i-1] == '\n')
		buff[i-1] = '\0';
}

const char* ReadInputString (void) {
	static char buff[1024];
	ReadInputString(buff, 1024);
	return buff;
}

const std::string MakeString (const char* format, ...) {
	va_list args;
	va_start(args, format);
	static char constructStr[1024]; 
	vsprintf(constructStr, format, args);
	va_end(args);
	return constructStr;
}

void Error (const char* msg) {
	NOTIFY_PROGRAMERROR(msg);
	CLEANUP_FUNCTION();
}

///////////////////////////////////////////////////////////////////
// IntObject class
// Used to support consts and vars
//
class IntObject {

	private:
	unsigned		MemCell;
	std::string		name;
	static unsigned currMemCell;

	public:
	IntObject (const char* s = "<array>") : 
		name(s), MemCell(currMemCell++) { SetValue(0); }

	IntObject (int x, const char* s = "<temp>") : 
		name(s), MemCell(currMemCell++) { SetValue(x); }

	IntObject (const IntObject& x, const char* s = "<temp>") : 
		name(s), MemCell(x.MemCell) {}

	~IntObject () {}

	//*****************************************************

	void		SetValue (int x);
	int			GetValue (void) const;

	void		SetName (const std::string& s) { name = s; }
	const char* GetName (void) const 
					{ return name.c_str(); }

	//*****************************************************
	// Arithmetic, Relational and Binary operators
	// implemented for IntObjects and native integers
	// Division not implemented as there is no DIV command

	#define INTOBJECT_OPERATOR(_op)\
	const int operator _op (const IntObject& x) const	\
		{ return(GetValue() _op x.GetValue()); }\
	\
	const int operator _op (int x) const \
		{ return(GetValue() _op x); }

	INTOBJECT_OPERATOR(+)
	INTOBJECT_OPERATOR(-)
	INTOBJECT_OPERATOR(*)
	INTOBJECT_OPERATOR(==)
	INTOBJECT_OPERATOR(!=)
	INTOBJECT_OPERATOR(>)
	INTOBJECT_OPERATOR(>=)
	INTOBJECT_OPERATOR(<)
	INTOBJECT_OPERATOR(<=)
	INTOBJECT_OPERATOR(&)
	INTOBJECT_OPERATOR(|)
	INTOBJECT_OPERATOR(^)
	INTOBJECT_OPERATOR(>>)
	INTOBJECT_OPERATOR(<<)
};

unsigned IntObject::currMemCell = 0;

///////////////////////////////////////////////////////////////////

class Array {

	private:
	std::string name;
	std::vector<IntObject>* array;
	
	class SetName : public std::binary_function<IntObject, std::string, bool> {
		static unsigned N;
		public:
		bool operator()(const IntObject& obj, const std::string& name) const {
				const_cast<IntObject&>(obj).SetName(MakeString("%s[%d]", name.c_str(), N++)); 
				return true; 
		}
		SetName (void) { N = 0; }
	};

	void Make (unsigned n) {
		array = new std::vector<IntObject>(n); 
		std::for_each(
			array->begin(),
			array->end(),
			std::bind2nd(SetName(), name)
		);		
	}

	public:
	IntObject& operator[](unsigned index) 
		{ assert(index < array->size()); return (*array)[index]; }

	Array (const char* s, unsigned N) : name(s) { Make(N); }
	Array (const char* s, const IntObject& N) : name(s) { assert(N.GetValue() > 0); Make(N.GetValue()); }
	~Array() { delete array; }
};

unsigned Array::SetName::N;

///////////////////////////////////////////////////////////////////
// Registers
//
class Reg {
	private:
	int						content;
	unsigned short			register_number;
	static unsigned short	register_count;
	
	public:
	Reg (void)	: content(0), register_number(register_count++) 
		{ assert(register_count <= TOTAL_REGISTERS); }

	Reg (int _content) : content(_content), register_number(TOTAL_REGISTERS) {}

	Reg (const Reg& x) : content(x.content), register_number(TOTAL_REGISTERS) {}

	virtual ~Reg () {}
	
	//*****************************************************

	int				GetContent (void) const 
						{ return content; }

	virtual void	SetContent (int value) 
						{ content = value; }

	unsigned short	GetRegNo (void) const 
						{ return register_number; }
	
	virtual const Reg operator= (const Reg& x);

	//*****************************************************
	// Arithmetic, Relational and Binary operators
	// implemented for Regs, IntObjects and native integers
	// Division not implemented as there is no DIV command

	#define REG_OPERATOR(_op)\
	const Reg operator _op (const Reg& x) const	\
		{ return Reg(content _op x.content); }\
	\
	const Reg operator _op (int x) const \
		{ return Reg(content _op x); }\
	\
	const Reg operator _op (IntObject x) const \
		{ return Reg(content _op x.GetValue()); }

	REG_OPERATOR(+)
	REG_OPERATOR(-)
	REG_OPERATOR(*)
	REG_OPERATOR(==)
	REG_OPERATOR(!=)
	REG_OPERATOR(>)
	REG_OPERATOR(>=)
	REG_OPERATOR(<)
	REG_OPERATOR(<=)
	REG_OPERATOR(&)
	REG_OPERATOR(|)
	REG_OPERATOR(^)
	REG_OPERATOR(>>)
	REG_OPERATOR(<<)
};

unsigned short Reg::register_count = 0;

class ZeroReg : public Reg {

	public:
	ZeroReg(void){}
	~ZeroReg(){}

	void SetContent (int value) { /* No warning as this is called only explicilty */ }

	const Reg operator= (const Reg& x) {
		OUTPUT_FUNCTION("WARNING : Assignment to R0 is ignored.\n");
		return *this;
	}
};


///////////////////////////////////////////////////////////////////////
// Native Reg and IntObject operators
//
#define INTOBJECT_NATIVE_OPERATOR(_op)\
		const int operator _op (int x, const IntObject& y) {\
			IntObject temp(x);\
			return (temp.operator _op(y));\
		}\

INTOBJECT_NATIVE_OPERATOR(+)
INTOBJECT_NATIVE_OPERATOR(-)
INTOBJECT_NATIVE_OPERATOR(==)
INTOBJECT_NATIVE_OPERATOR(!=)
INTOBJECT_NATIVE_OPERATOR(>)
INTOBJECT_NATIVE_OPERATOR(>=)
INTOBJECT_NATIVE_OPERATOR(<)
INTOBJECT_NATIVE_OPERATOR(<=)
INTOBJECT_NATIVE_OPERATOR(&)
INTOBJECT_NATIVE_OPERATOR(|)
INTOBJECT_NATIVE_OPERATOR(^)
INTOBJECT_NATIVE_OPERATOR(>>)
INTOBJECT_NATIVE_OPERATOR(<<)


#define REG_NATIVE_OPERATOR(_op)\
		const Reg operator _op (int x, const Reg& y) {\
			Reg temp(x);\
			return (temp.operator _op(y));\
		}\
		\
		const Reg operator _op (IntObject x, const Reg& y) {\
			Reg temp(x.GetValue());\
			return (temp.operator _op(y));\
		}

REG_NATIVE_OPERATOR(+)
REG_NATIVE_OPERATOR(-)
REG_NATIVE_OPERATOR(==)
REG_NATIVE_OPERATOR(!=)
REG_NATIVE_OPERATOR(>)
REG_NATIVE_OPERATOR(>=)
REG_NATIVE_OPERATOR(<)
REG_NATIVE_OPERATOR(<=)
REG_NATIVE_OPERATOR(&)
REG_NATIVE_OPERATOR(|)
REG_NATIVE_OPERATOR(^)
REG_NATIVE_OPERATOR(>>)
REG_NATIVE_OPERATOR(<<)


///////////////////////////////////////////////////////////////////
// Memory
//
class MemCell {
	unsigned char content;

	public:
		//*****************************************************
		// Constructors / Destructor

		MemCell (void) : content(0) {}

		MemCell (const MemCell& x) : content(x.content) {}

		~MemCell (){}
		
		//*****************************************************

		unsigned char GetContent (void) const { return content; }

		void SetContent (unsigned char value) { content = value; }	//Used only explicitly at SetMem

		const MemCell operator= (unsigned char x);
};


///////////////////////////////////////////////////////////////////
// Jet Singleton
//
class Jet {
	public:
		Jet (void)	{}
		~Jet ()		{}
		
		ZeroReg zero;
		Reg regs[TOTAL_REGISTERS - 1];
		MemCell mem[MEMORY_SIZE >> 2][4];	//in bytes

		//***************************************************

		unsigned GetMemCellNo (MemCell* memCell) const
			{ return memCell - &mem[0][0];}			

		unsigned GetRegNo (const Reg& reg) {
			unsigned ret = (&reg == &zero) ? 0 : (&reg - regs + 1);
			assert(ret == reg.GetRegNo());
			return ret;
		}

		// These are used in order to have branches that
		// can have both operands regs and/or numbers, consts, vars
		unsigned GetRegNo (int x)		const { return x; }
		unsigned GetRegNo (IntObject x) const { return x.GetValue(); }

		//***************************************************

		void print_regs (unsigned line) const {
			OUTPUT_FUNCTION("R0 = 0\n");
			for(unsigned i = 1; i < TOTAL_REGISTERS; ++i)
				OUTPUT_FUNCTION("At line %d, R%d = %d\n", line, i, regs[i - 1].GetContent());
		}

		void print_reg (unsigned line, Reg& reg)
			{ OUTPUT_FUNCTION("At line %d, R%d = %d\n", line, GetRegNo(reg), reg.GetContent()); }
};

static Jet jet;

/////////////////////////
//Register definitions
//
#define R0 jet.zero
#define R1 jet.regs[0]
#define R2 jet.regs[1]
#define R3 jet.regs[2]
#define R4 jet.regs[3]
#define R5 jet.regs[4]
#define R6 jet.regs[5]
#define R7 jet.regs[6]
#define R8 jet.regs[7]
#define R9 jet.regs[8]
#define R10 jet.regs[9]
#define R11 jet.regs[10]
#define R12 jet.regs[11]
#define R13 jet.regs[12]
#define R14 jet.regs[13]
#define R15 jet.regs[14]
#define R16 jet.regs[15]
#define R17 jet.regs[16]
#define R18 jet.regs[17]
#define R19 jet.regs[18]
#define R20 jet.regs[19]
#define R21 jet.regs[20]
#define R22 jet.regs[21]
#define R23 jet.regs[22]
#define R24 jet.regs[23]
#define R25 jet.regs[24]
#define R26 jet.regs[25]
#define R27 jet.regs[26]
#define R28 jet.regs[27]
#define R29 jet.regs[28]
#define R30 jet.regs[29]
#define R31 jet.regs[30]


#define NUM_OF_REG(reg)	jet.GetRegNo(reg)
#define MEMORY jet.mem

///////////////////////////////////////////////////////////////////////////////
// These functions need to be implemented here, after the jet declaration
//
const Reg Reg::operator= (const Reg& x) {
	assert(register_number < TOTAL_REGISTERS);
	content = x.content;

	NOTIFY_REGASSIGN(NUM_OF_REG(*this), content);
	return *this;
}

const MemCell MemCell::operator= (unsigned char x) {
	content = x;
	NOTIFY_MEMASSIGN(jet.GetMemCellNo(this), x);
	return *this;
}

//************************************************************************

void IntObject::SetValue (int x) {
	if(MemCell >= MEMORY_SIZE >> 2)
		Error("Out of Memory!\n");

	MEMORY[MemCell][3] = (unsigned char) (x >> 24);
	MEMORY[MemCell][2] = (unsigned char) (x >> 16);
	MEMORY[MemCell][1] = (unsigned char) (x >> 8);
	MEMORY[MemCell][0] = (unsigned char)  x;
}

int IntObject::GetValue (void) const {
	return	(((int) MEMORY[MemCell][3].GetContent()) << 24) + 
			(((int) MEMORY[MemCell][2].GetContent()) << 16) + 
			(((int) MEMORY[MemCell][1].GetContent()) << 8) + 
			  (int) MEMORY[MemCell][0].GetContent();
}

///////////////////////////////////////////////////////////////////
// R-format instructions
// (Instructions that have operands (Reg, Reg, Reg) )
//
#define R_FORMAT(_name, _op)	\
	void _##_name (unsigned line, Reg& Ra, Reg& Rb, Reg& Rc) {\
		NOTIFY_RFORMAT(#_name, NUM_OF_REG(Ra), NUM_OF_REG(Rb), NUM_OF_REG(Rc), line);\
		Rc = Ra _op Rb;\
	}

	R_FORMAT(ADD, +)
	R_FORMAT(SUB, -)
	R_FORMAT(MUL, *)
	R_FORMAT(CMEQ, ==)	//Compare equal
	R_FORMAT(CMNE, !=)	//Compare not equal
	R_FORMAT(CMGE, >=)	//Compare greater or equal
	R_FORMAT(CMLT, <)	//Compare less than
	R_FORMAT(AND, &)
	R_FORMAT(OR, |)
	R_FORMAT(XOR, ^)
	R_FORMAT(SHRU, >>)	//Shift right unsigned - shift is by default unsigned
	R_FORMAT(SHL, <<)	//Shift left


	//SHRS - Shift right signed
	void _SHRS (unsigned line, Reg& Ra, Reg& Rb, Reg& Rc) {
		NOTIFY_RFORMAT("SHRS", NUM_OF_REG(Ra), NUM_OF_REG(Rb), NUM_OF_REG(Rc), line);
		Rc = Ra >> Rb;
		
		if(Rc.GetContent() & (2 << (BITS_PER_REGISTER - Rb.GetContent())))
			for(int i = BITS_PER_REGISTER - Rb.GetContent(); i < BITS_PER_REGISTER; ++i)
				Rc = Rc | (2 << i);
	}

	//GCP - Guarded copy
	void _GCP (unsigned line, Reg& Ra, Reg& Rb, Reg& Rc) { 
		NOTIFY_RFORMAT("GCP", NUM_OF_REG(Ra), NUM_OF_REG(Rb), NUM_OF_REG(Rc), line);
		if(Ra.GetContent()) Rc = Rb; 
	}

//************************************************************************************
// Use these macros to get line information
//

#define	LABEL(id)	); id: (0

#define ADD		); _ADD(__LINE__,
#define SUB		); _SUB(__LINE__,
#define MUL		); _MUL(__LINE__,
#define CMEQ	); _CMEQ(__LINE__,
#define CMNE	); _CMNE(__LINE__,
#define CMGE	); _CMGE(__LINE__,
#define CMLT	); _CMLT(__LINE__,
#define CMLE	); _CMLE(__LINE__,
#define AND		); _AND(__LINE__,
#define OR		); _OR(__LINE__,
#define XOR		); _XOR(__LINE__,
#define SHRU	); _SHRU(__LINE__,
#define SHL		); _SHL(__LINE__,
#define SHRS	); _SHRS(__LINE__,	
#define GCP		); _GCP(__LINE__,

///////////////////////////////////////////////////////////////////
// I-format instructions
// (Instructions that have operands (Reg, Reg, Imm13), Imm13 = immediate of 13 bits)
//
#define FIX_BIT_PRECISION(x, bits)	\
		((x << (BITS_PER_REGISTER - bits)) >> (BITS_PER_REGISTER - bits))

#define I_FORMAT(_name, _op)	\
	void _##_name (unsigned line, Reg& Ra, Reg& Rb, int Imm13) {\
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);\
		NOTIFY_IFORMAT(#_name, NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);\
		Rb = Ra _op Imm13;\
	}\
	/*This is for supporting I-format instructions with IntObjects(consts, vars, array elements)*/\
	void _##_name (unsigned line, Reg& Ra, Reg& Rb, IntObject Imm13)\
		{ _##_name(line, Ra, Rb, Imm13.GetValue()); }

	I_FORMAT(ADDI, +)
	I_FORMAT(SUBI, -)
	I_FORMAT(MULI, *)
	I_FORMAT(CMEQI, ==)	//Compare equal immediate
	I_FORMAT(CMNEI, !=)	//Compare not equal immediate
	I_FORMAT(CMGEI, >=)	//Compare greater or equal immediate
	I_FORMAT(CMLTI, <)	//Compare less than immediate
	I_FORMAT(CMLEI, <=)	//Compare less or equal immediate
	I_FORMAT(ANDI, &)
	I_FORMAT(ORI, |)
	I_FORMAT(XORI, ^)
	R_FORMAT(SHRUI, >>)	//Shift right unsigned immediate - shift is by default unsigned
	R_FORMAT(SHLI, <<)	//Shift left immediate


//*************************************************************************
// Add IntObject support for i-format instructions
//
#define	SUPPORT_IFORMAT_INTOBJECTS(_name)\
	void _##_name (unsigned line, Reg& Ra, Reg& Rb, IntObject Imm13)\
		{ _##_name(line, Ra, Rb, Imm13.GetValue()); }


	//SHRSI - Shift right signed immediate
	void _SHRSI (unsigned line, Reg& Ra, Reg& Rb, int Imm13) {
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);
		NOTIFY_IFORMAT("SHRSI", NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);
		Rb = Ra >> Imm13;
		if(Rb.GetContent() & (2 << (BITS_PER_REGISTER - Imm13)))
			for(int i = BITS_PER_REGISTER - Imm13; i < BITS_PER_REGISTER; ++i)
				Rb = Rb | (2 << i);
	}
	
	SUPPORT_IFORMAT_INTOBJECTS(SHRSI)

	
	//SUBII - Sub inverse immediate
	void _SUBII (unsigned line, Reg& Ra, Reg& Rb, int Imm13) { 
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);
		NOTIFY_IFORMAT("SUBII", NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);
		Rb = Imm13 - Ra; 
	}

	SUPPORT_IFORMAT_INTOBJECTS(SUBII)

	
	//***********************************
	//memory instructions

	//LDW - Load word
	void _LDW (unsigned line, Reg& Ra, Reg& Rb, int Imm13) { 
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);
		NOTIFY_IFORMAT("LDW", NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);
		int tmp = Ra.GetContent() + Imm13;
		if(tmp >= MEMORY_SIZE >> 2)
			Error("LDW : Index of memory out of bounds.");
		Rb = (((int) MEMORY[tmp][3].GetContent()) << 24) + 
		     (((int) MEMORY[tmp][2].GetContent()) << 16) + 
			 (((int) MEMORY[tmp][1].GetContent()) << 8) + 
			   (int) MEMORY[tmp][0].GetContent();
	}

	SUPPORT_IFORMAT_INTOBJECTS(LDW)

	//STW - Store word
	void _STW (unsigned line, Reg& Ra, Reg& Rb, int Imm13) {
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);
		NOTIFY_IFORMAT("STW", NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);
		int tmp1 = Ra.GetContent() + Imm13;
		if(tmp1 >= MEMORY_SIZE >> 2)
			Error("STW : Index of memory out of bounds.");
		int tmp2 = Rb.GetContent();

		MEMORY[tmp1][3] = (unsigned char) (tmp2 >> 24);
		MEMORY[tmp1][2] = (unsigned char) (tmp2 >> 16);
		MEMORY[tmp1][1] = (unsigned char) (tmp2 >> 8);
		MEMORY[tmp1][0] = (unsigned char)  tmp2;
	}

	SUPPORT_IFORMAT_INTOBJECTS(STW)

	//***************************************************************
	
	//LDBU - Load byte unsigned
	void _LDBU (unsigned line, Reg& Ra, Reg& Rb, int Imm13) {
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);
		NOTIFY_IFORMAT("LDBU", NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);
		int tmp = Ra.GetContent() + Imm13;
		if(tmp >= MEMORY_SIZE)
			Error("LDBU : Index of memory out of bounds.");
		Rb = (int) MEMORY[tmp >> 2][tmp & 3].GetContent();
	}

	SUPPORT_IFORMAT_INTOBJECTS(LDBU)


	//LDBS - Load byte signed
	void _LDBS (unsigned line, Reg& Ra, Reg& Rb, int Imm13) {
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);
		NOTIFY_IFORMAT("LDBS", NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);
		int tmp = Ra.GetContent() + Imm13;
		if(tmp >= MEMORY_SIZE)
			Error("LDBS : Index of memory out of bounds.");
		Rb = (int) MEMORY[tmp >> 2][tmp & 3].GetContent();
		if(Rb.GetContent() & (2 << 7))
			for(int i = 7; i < BITS_PER_REGISTER; ++i)
				Rb = Rb | (2 << i);
	}

	SUPPORT_IFORMAT_INTOBJECTS(LDBS)


	//STB - Store byte
	void _STB (unsigned line, Reg& Ra, Reg& Rb, int Imm13) {
		Imm13 = FIX_BIT_PRECISION(Imm13, 13);
		NOTIFY_IFORMAT("STB", NUM_OF_REG(Ra), NUM_OF_REG(Rb), Imm13, line);
		int tmp = Ra.GetContent() + Imm13;
		if(tmp >= MEMORY_SIZE)
			Error("STB : Index of memory out of bounds.");
		MEMORY[tmp >> 2][tmp & 3] = (unsigned char) Rb.GetContent();
	}

	SUPPORT_IFORMAT_INTOBJECTS(STB)

//************************************************************************************
///////////////////////////////////////////
// Use these macros to get line information
//
#define ADDI	); _ADDI(__LINE__,
#define SUBI	); _SUBI(__LINE__,
#define MULI	); _MULI(__LINE__,
#define CMEQI	); _CMEQI(__LINE__,
#define CMNEI	); _CMNEI(__LINE__,
#define CMGEI	); _CMGEI(__LINE__,
#define CMLTI	); _CMLTI(__LINE__,
#define CMLEI	); _CMLEI(__LINE__,
#define ANDI	); _ANDI(__LINE__,
#define ORI		); _ORI(__LINE__,
#define XORI	); _XORI(__LINE__,
#define SHRUI	); _SHRUI(__LINE__,
#define SHLI	); _SHLI(__LINE__,
#define SHRSI	); _SHRSI(__LINE__,
#define SUBII	); _SUBII(__LINE__,
#define LDW		); _LDW(__LINE__,
#define STW		); _STW(__LINE__,
#define LDBU	); _LDBU(__LINE__,
#define LDBS	); _LDBS(__LINE__,
#define STB		); _STB(__LINE__,

///////////////////////////////////////////////////////////////////
// Branches
//
#define REG_REG	0
#define REG_INT	1
#define INT_REG	2
#define INT_INT	3

unsigned char branch_id_from_arguments(const Reg& a, const Reg& b)	
	{ return REG_REG; }
unsigned char branch_id_from_arguments(const Reg& a, int b)	
	{ return REG_INT; }
unsigned char branch_id_from_arguments(int a, const Reg& b)	
	{ return INT_REG; }
unsigned char branch_id_from_arguments(int a, int b)	
	{ return INT_INT; }

unsigned char branch_id_from_arguments(IntObject a, const Reg& b)	
	{ return INT_REG; }
unsigned char branch_id_from_arguments(const Reg& a, const IntObject& b)	
	{ return REG_INT; }
unsigned char branch_id_from_arguments(IntObject a, int b)		
	{ return INT_INT; }
unsigned char branch_id_from_arguments(int a, const IntObject& b)		
	{ return INT_INT; }

#define	BRANCH_REGOP(name, op, str) \
template <class T> \
inline bool name (unsigned line, const Reg& r, const T& arg, const char* label) { \
		NOTIFY_BRANCH( \
				str, \
				branch_id_from_arguments(r, arg), \
				NUM_OF_REG(r), \
				NUM_OF_REG(arg), \
				label, \
				line \
			); \
		Reg tmp(r op arg); \
		return !!tmp.GetContent(); \
}

BRANCH_REGOP(breq, ==, "BREQ")
BRANCH_REGOP(brne, !=, "BRNE")
BRANCH_REGOP(brge, >=, "BRGE")
BRANCH_REGOP(brlt, <=, "BRLT")

#define BREQ	); if (breq(__LINE__,
#define BRNE	); if (brne(__LINE__,
#define BRGE	); if (brge(__LINE__,
#define BRLT	); if (brlt(__LINE__,
#define	TO(id)		#id)) goto id; (0

///////////////////////////////////////////////////////////////////
// L-format instructions
// (Instructions that have operands (Reg, Imm19) , Imm19 = Immediate of 19 bits)
//

// Add IntObject support for I-format instructions
#define	SUPPORT_LFORMAT_INTOBJECTS(_name)\
		void _##_name (unsigned line, Reg& Ra, IntObject Imm19)\
			{ _##_name(line, Ra, Imm19.GetValue()); }


	//SETHI - Set High bits
	void _SETHI (unsigned line, Reg& Ra, int Imm19) { 
		Imm19 = FIX_BIT_PRECISION(Imm19, 19);
		NOTIFY_LFORMAT("SETHI", NUM_OF_REG(Ra), Imm19, line);
		Ra = Imm19 << 13; 
	}

	SUPPORT_LFORMAT_INTOBJECTS(SETHI)
	
///////////////////////////////////////////////////////////////////
// Direct jump.
//
inline bool jump (unsigned line, const char* label) {
	NOTIFY_JUMP("JUMP", label, line);
	return true;	// Unused.
}

#define JUMP	); if (jump(__LINE__,


///////////////////////////////////////////////////////////////////
// Using the visualizer to change the contents of regs or memory
//
#ifdef SUPPLY_VISUALIZER_ACCESS_FUNCTIONS
	void SetJetReg (unsigned num, int value) {
		assert(num < TOTAL_REGISTERS);
		if(num) // as not to assign a value to R0
			jet.regs[num - 1].SetContent(value);
	}

	void SetJetMem (unsigned pos, unsigned char value)
		{ MEMORY[pos / 4][pos % 4].SetContent(value); }
#endif

///////////////////////////////////////////////////////////////////
// Definitions and functions for pseudo-instructions
//
#define CONSTANT(name, value)	const IntObject name(value, #name);
#define VARIABLE(name, value)	IntObject name(value, #name);
#define ARRAY(name, size)		Array name(#name, size);

//******************************************************************

#define	SETVAR	); _SETVAR(__LINE__,

void _SETVAR (unsigned line, IntObject& var, Reg& reg)		{ var.SetValue(reg.GetContent()); }
void _SETVAR (unsigned line, IntObject& var, IntObject x)	{ var.SetValue(x.GetValue()); }
void _SETVAR (unsigned line, IntObject& var, int x)			{ var.SetValue(x); }

//******************************************************************

int GetInput (void) {
	int x;
	while(sscanf(ReadInputString(), "%d", &x) != 1)
		OUTPUT_FUNCTION("Please insert a value!\n");
	return x;
}

#define	INPUT	); _INPUT(__LINE__,

void _INPUT(unsigned line, Reg& reg)		{ reg = GetInput(); }
void _INPUT(unsigned line, IntObject &var)	{ var.SetValue(GetInput()); }

//******************************************************************`

#define OUTPUT	); _OUTPUT(__LINE__,

void _OUTPUT (unsigned line, Reg& reg)
	{ OUTPUT_FUNCTION("R%d = %d\n", NUM_OF_REG(reg), reg.GetContent()); }

void _OUTPUT (unsigned line, const char *str)		
	{ OUTPUT_FUNCTION("%s\n", str); }

void _OUTPUT (unsigned line, IntObject& var)	
	{ OUTPUT_FUNCTION("%s = %d\n", var.GetName(), var.GetValue()); }

#define PRINT_REGS	); jet.print_regs(__LINE__
#define PRINT_REG	); jet.print_reg(__LINE__,


///////////////////////////////////////////////////////////////////
// Defines and function for main
//
#define INSTANTIATION_TRAPPER(_class, call)	\
	class _class##PreInstantiation\
	{\
		public:\
			_class##PreInstantiation(void)\
			{\
				call;\
			}\
	};\
	_class##PreInstantiation _instantiation;
	

INSTANTIATION_TRAPPER(_Program, NOTIFY_PROGRAMSTART())

#define PROGRAM_BEGIN	\
	int main (int argc, char **argv) { (0

#define PROGRAM_END			\
		);					\
		NOTIFY_PROGRAMEND();\
		return 0;			\
	}

#endif	// Do not add stuff beyond this point.
