// VisualiserSuper.h
// Super-class for a JET execution visualizer.
// Virtual functions correspond to what has to be 
// implemented per specific visualizer.
// A. Savidis, Y.Lilis.
// November, 2005.
// Project: JET Suite (Execution visualizer, Savidis, 2005).
//

#if	defined(HAS_VISUALIZER) && (defined(CONSOLE_VISUALIZER) ||defined(GRAPHICAL_VISUALIZER))

#ifndef	VISUALISERSUPER_H
#define	VISUALISERSUPER_H

#pragma warning(disable:4786)

#include <map>
#include <string>
#include <stdarg.h>
#include "Configurator.h"
#include "ExecutionEvents.h"

////////////////////////////////////////////////////////////////////
// Externalised by the Jet header file
//
extern const std::string	MakeString (const char* format,...);

////////////////////////////////////////////////////////////////////

class VisualiserSuper {

	protected:
	std::map<std::string, bool>	notifyingEvents;	// Which execution events "notify".
	std::map<std::string, bool>	steppingEvents;		// Which execution events "step".

	// The following must be declared and defined by the derived
	// class as needed. Derived classes are mutually exclusive. The static
	// variables should never be explicitly initialized. The GetVisualiser()
	// function should create and assign upon the first call the derived visualiser.
	//
	static bool				cleanedUp;
	static VisualiserSuper* visualiser;

	VisualiserSuper (void)	{ Initialise(); visualiser = this; }
	virtual ~VisualiserSuper()	{ CleanUp(); visualiser = (VisualiserSuper*) 0; }

	public:

	static VisualiserSuper* GetVisualiser (void);

	/////////////////////////////////////////////////////////////////
	// Those are the only functions that the particular visualizer
	// MUST implement. All of them are pure virtual functions.
	//
	virtual void Message (const char* str) const = 0;	// An execution trace message.

	virtual void MemCellRefresh (unsigned MemCell, int value) = 0;

	virtual void RegRefresh (unsigned reg, int value) = 0;

	virtual void StepWait (void) const = 0;

	/////////////////////////////////////////////////////////////////
	// Refinement of those functions allows thead-enabled versions.
	//
	virtual bool IsNotifyingEvent (const std::string& name)
		{ return notifyingEvents[name]; }

	virtual void SetNotifyingEvent (const std::string& name, bool enabled)
		{ notifyingEvents[name] = enabled; }

	virtual bool IsSteppingEvent (const std::string& name)
		{ return steppingEvents[name]; }

	virtual void SetSteppingEvent (const std::string& name, bool enabled)
		{ steppingEvents[name] = enabled; }

	protected:

	/////////////////////////////////////////////////////////////////

#define	FUNCMACRO_NOTIFIER(x, declargs, callargs)			\
	static void On##x declargs {							\
		if (GetVisualiser())								\
			if(GetVisualiser()->IsNotifyingEvent(#x)) {		\
				GetVisualiser()->x callargs;				\
				if(GetVisualiser()->IsSteppingEvent(#x))	\
					GetVisualiser()->StepWait();			\
			}												\
	}

#define	FUNCMACRO_MESSAGER(x, declargs, callargs)	\
	void x declargs {								\
		Message(									\
			MakeString callargs.c_str()				\
		);											\
	}

#define	MSG(x)	Configurator::On##x

#define	FUNCMACRO_META(x, declargs, callargs)							\
	FUNCMACRO_MESSAGER(x, (declargs), (Configurator::On##x, callargs))	\
	FUNCMACRO_NOTIFIER(x, (declargs), (callargs))

	//************************

#define	RFORMAT_DECLARGS	const char* name, unsigned ra, unsigned rb, unsigned rc, unsigned line
#define	RFORMAT_CALLARGS	name, ra, rb, rc, line

	FUNCMACRO_META(RFormat, RFORMAT_DECLARGS, RFORMAT_CALLARGS)

#define	IFORMAT_DECLARGS	const char* name, unsigned ra, unsigned rb, int imm13, unsigned line
#define	IFORMAT_CALLARGS	name, ra, rb, imm13, line

	FUNCMACRO_META(IFormat, IFORMAT_DECLARGS, IFORMAT_CALLARGS)

#define	LFORMAT_DECLARGS	const char* name, unsigned ra, int imm19, unsigned line
#define	LFORMAT_CALLARGS	name, ra, imm19, line

	FUNCMACRO_META(LFormat, LFORMAT_DECLARGS, LFORMAT_CALLARGS)

#define	JUMP_DECLARGS	const char* name, const char* label, unsigned line
#define	JUMP_CALLARGS	name, label, line

	FUNCMACRO_META(Jump, JUMP_DECLARGS, JUMP_CALLARGS)

	//************************

#define	BRANCH_DECLARGS	const char* name, unsigned char mask, unsigned ra, unsigned rb, const char* label, unsigned line
#define	BRANCH_CALLARGS	name, mask, ra, rb, label, line

	void Branch (BRANCH_DECLARGS) {
		char* str[2] = {"R", "R"};

		if(mask & 1)
			str[0] = "";
		if(mask & 2)
			str[1] = "";

		Message(
			MakeString(
				MSG(Branch), 
				name, 
				str[1], 
				ra, 
				str[0], 
				rb, 
				label, 
				line
			).c_str()
		);
	}

	FUNCMACRO_NOTIFIER(Branch, (BRANCH_DECLARGS), (BRANCH_CALLARGS))

	//************************

#define ASSIGN_DECLARGS	unsigned pos, int value		// used both for Reg and Mem
#define ASSIGN_CALLARGS	pos, value					// used both for Reg and Mem

	void RegAssign (ASSIGN_DECLARGS)
		{ RegRefresh(ASSIGN_CALLARGS); }
	
	FUNCMACRO_NOTIFIER(RegAssign, (ASSIGN_DECLARGS), (ASSIGN_CALLARGS))


	void MemAssign (ASSIGN_DECLARGS)
		{ MemCellRefresh(ASSIGN_CALLARGS); }

	FUNCMACRO_NOTIFIER(MemAssign, (ASSIGN_DECLARGS), (ASSIGN_CALLARGS))

/////////////////////////////////////////////////////////////////

#define INIT_FUNC(x)								\
			ExecutionEvents::SetOn##x(On##x);		\
			SetNotifyingEvent(#x, true);			\
			SetSteppingEvent(#x, false);

	void Initialise (void) {
		INIT_FUNC(RegAssign)
		INIT_FUNC(MemAssign)
		INIT_FUNC(RFormat)
		INIT_FUNC(IFormat)
		INIT_FUNC(LFormat)
		INIT_FUNC(Branch)
		INIT_FUNC(Jump)
	}

	void CleanUp (void) {
		notifyingEvents.clear();
		steppingEvents.clear();
	}
};

//********************************

class VisualiserSuper_Init {
	private:
	static bool done;

	public:
	VisualiserSuper_Init (void);
};

static VisualiserSuper_Init visualiserInit;

#define	VISUALISER_INIT									\
	bool VisualiserSuper_Init::done;					\
	VisualiserSuper_Init::VisualiserSuper_Init (void) {	\
		if (!done) {									\
			VisualiserSuper::GetVisualiser();			\
			assert(VisualiserSuper::GetVisualiser());	\
		}												\
		else											\
			done = true;								\
	}

////////////////////////////////////////////////////////////////
// The linkage to mutually exclusive visualisers.
// They need declare the visualiser variable and implement
// the visualiser instance accessor and creation function. We
// never initialise explicitly, but rely on zeroing at start-up.
// We need only put the macro with the derived class name as a 
// parameter in the visualiser implementation file.
//
#define	VISUALISER_LINKAGE(_class)							\
bool				VisualiserSuper::cleanedUp;				\
VisualiserSuper*	VisualiserSuper::visualiser;			\
VisualiserSuper* VisualiserSuper::GetVisualiser (void) {	\
	/* After clean-up we disable visualisations. */			\
	if (!visualiser && !cleanedUp)	{						\
		new _class;											\
		assert(visualiser);									\
	}														\
	return visualiser;										\
}															\
VISUALISER_INIT

///////////////////////////////////////////////////////

#endif	// Header file end point.

#endif	// HAS_VISUALIZER
