// GraphicalUI.cpp
// A GTK implementation of the JET Visualizer
// A. Savidis, Y.Lilis.
// November, 2005.
// Project: JET Suite (Execution visualizer, Savidis, 2005).
//

#if	defined(HAS_VISUALIZER) && defined(GRAPHICAL_VISUALIZER)

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <assert.h>
#include <string>
#include <string.h>
#include <errno.h>
#include "GraphicalUILinkage.h"
#include "DataMirror.h"
#include "Specs.h"
#include "Configurator.h"

//////////////////////////////////
// Visualizer GUI specific values
//
#define WINDOW_WIDTH		800
#define WINDOW_HEIGHT		600

#define IF_DESTROYED_EXIT()	do { if(!window) return; } while(false)

//////////////////////////////////////
// Function externalized from Servant
//
extern const std::string MakeString(const char *format, ...);

////////////////////////////////////////
// Functions found in this file. These functions 
// manage the Gtk Implementation of the UI. They
// are for private use and declared static.
//

#ifdef SUPPLY_VISUALIZER_ACCESS_FUNCTIONS
extern void SetJetReg(unsigned num, int value);
extern void SetJetMem(unsigned pos, unsigned char value);

static const char * RemoveSpaces (const char *str);
static void OnRegActivate (GtkEntry *entry, gpointer user_data);
static void OnMemActivate (GtkEntry *entry, gpointer user_data);
#endif

#ifdef LARGE_MEMORY
static void OnPreviousClicked (GtkButton *button, gpointer user_data);		//Called upon pressing "previous"
static void OnNextClicked (GtkButton *button, gpointer user_data);			//Called upon pressing "next"
#endif

static void OnStepClicked (GtkButton *button, gpointer user_data);			//Called upon pressing "step"

static void OnInputActivate (GtkEntry *entry, gpointer user_data);			//Called upon pressing return on Input entry

/////////////////////////////////////////////////////////////
// Memory and register display implementation and callbacks
//

namespace GTKDISPLAY {
static char *DisplayBin(int x, unsigned digits);							//Possible display functions. Memory can only
static char *DisplayDec(int x, unsigned digits);							//have bin or hex while regs can have all 3.
static char *DisplayHex(int x, unsigned digits);

static char * (*displayMem)(int x, unsigned digits) = DisplayBin;			//Points to the current mem display function
static char * (*displayRegs)(int x, unsigned digits) = DisplayDec;			//Points to the current reg display function

static void SetDisplay (unsigned char whichDisplay, unsigned char whatDisplay);	//Sets the display for either mem or reg

static void DisplayRegAction (GtkAction *action, GtkRadioAction *current);	//Called when changing the reg display option
static void DisplayMemAction (GtkAction *action, GtkRadioAction *current);	//Called when changing the mem display option
}

//////////////////////////////
// Defines for the displays
//
#define ZERO_BIN	"00000000"
#define ZERO_HEX	"00"

#define REGS	0
#define MEM		1
#define BIN		0
#define DEC		1
#define HEX		2

//////////////////////////////////////////////
// The Menu Bar implementation and callbacks
//
namespace GTKMENU {
static void Save (const char* name);									//Saves the whole program state to a file
static void StoreFilename (GtkWidget* widget, gpointer user_data);		//Gets the filename inserted in the appropriate dialog box
static void CreateFileSelection (void);									//

static void SetNotifyingState (GtkAction *action);						//Called when changing an notifying event
static void SetSteppingState (GtkAction *action);						//Called when changing a steping event

static void File (GtkAction* action);									//Menu->File event dispatcher
static void ActivateEmail (GtkAboutDialog* about, const gchar* link, gpointer data);	//Called upon clicking an author or documentor name
static void About (GtkAction* action, GtkWidget* window);				//Shows information about the Visualizer

static void Create (GtkWidget* window, GtkWidget* content);				//Creates the Menu Bar for the Window of the UI
}

//****************************

static void GtkEnd (void);												//Ends the Visualizer main loop

static void GtkStatus (const char* str);								//Sets the program Status Bar

static void* GtkMain (void *unused);									//Visualizer thread that handles pending GTK events

static void GtkSetCurrentPage (unsigned page);							//Sets the Visualizer notebook to this page

static const char* GtkTextViewGetText (GtkTextView* textview);			//GetTextView getText wrapper function

static void GtkTextViewAddText(GtkTextView *textview, const char *msg);	//Adds the given text to the given textview

static GtkWidget* GtkCreateWindow (void);								//Creates the window for the UI

static void RefreshAllRegisters (void);
static void RefreshAllMemory (void);

//////////////////////////////////////////////////
// GUI data needed globally. All of them are for
// private use and therefore declared static.
//
static GtkWidget*	memory[PAGE_SIZE >> 2][5];
static GtkWidget*	regs[TOTAL_REGISTERS][2];
static GtkWidget*	TraceText;
static GtkWidget*	OutputText;
static GtkWidget*	InputEntry;
static GtkWidget*	statusbar;
static GtkWidget*	window;

static bool	GtkDone = false;

static bool messageAlreadyShown = false;

static unsigned	CurrentPage = 0;

static char*	input = (char *) 0;
static bool		inputDone = false;

static const char*	SelectedFilename = (const char*) 0;

///////////////////////////////////////////////////////////////////
// Mutexes and conditions for mutual exclusion and synchronization
//
static GMutex*	step_mutex;
static GMutex*	wait_mutex;
static GMutex*	visual_mutex;
static GCond*	step_cond;
static GCond*	wait_cond;
static GCond*	visual_cond;

//////////////////////////////////////////////////
// GRAPHICALUI namespace implementation functions
//


////////////////////////////////////////////
// The initialization and cleanup of the UI
//

void GRAPHICALUI::Create (void) {
	GtkDone = false;
	if(!g_thread_supported ()) {
		g_thread_init (NULL);
		gdk_threads_init();
	}
	
	g_thread_create(GtkMain, NULL, 1, NULL);

	step_mutex = g_mutex_new();
	wait_mutex = g_mutex_new();
	visual_mutex = g_mutex_new();
	step_cond = g_cond_new();
	wait_cond = g_cond_new();
	visual_cond = g_cond_new();

	g_mutex_lock(wait_mutex);
	g_cond_wait(wait_cond, wait_mutex);
	g_mutex_unlock(wait_mutex);
}


void GRAPHICALUI::Destroy (void){ 
  IF_DESTROYED_EXIT();
  GtkEnd(); 
}


///////////////////////////////////////////////////////////////
// Functions with which the main thread interacts with the UI.
//

void GRAPHICALUI::MemoryRefresh (unsigned index) {
	IF_DESTROYED_EXIT();
  
	// Not current page so don't change anything at screen
	//
	if(index / PAGE_SIZE != CurrentPage)	
		return;

	int content = 0;
		
	DataMirror::GetSingleton()->GrantMemAccess();
	if(DataMirror::GetSingleton()->IsMemOccupied(index))
		content = DataMirror::GetSingleton()->GetMemContent(index);
	DataMirror::GetSingleton()->ReleaseMemAccess();
	
	char *text = GTKDISPLAY::displayMem(content, 8);

	if(!GtkDone)
		gtk_entry_set_text(GTK_ENTRY(memory[index >> 2][4 - (index & 3)]), text);
}

void GRAPHICALUI::RegisterRefresh (unsigned reg) {
	IF_DESTROYED_EXIT();
  
	int content = 0;
		
	DataMirror::GetSingleton()->GrantRegAccess();
	if(DataMirror::GetSingleton()->IsRegOccupied(reg))
		content = DataMirror::GetSingleton()->GetRegContent(reg);
	DataMirror::GetSingleton()->ReleaseRegAccess();
	
	char *text = GTKDISPLAY::displayRegs(content, 32);

	if(!GtkDone)
		gtk_entry_set_text(GTK_ENTRY(regs[reg][1]), text);
}


//****************************

void GRAPHICALUI::Finished (void) {
	IF_DESTROYED_EXIT();
	GtkStatus("Program finished: press 'Step' to exit");
	messageAlreadyShown = true;
}

void GRAPHICALUI::Started (void) {
	IF_DESTROYED_EXIT();
	GtkStatus("Program started: press 'Step' to begin");
	messageAlreadyShown = true;
}

void GRAPHICALUI::StepWait (void) {
	IF_DESTROYED_EXIT();
	
	if(!messageAlreadyShown)
		GtkStatus("Stepping into event. Press 'Step' to continue");
	else
		messageAlreadyShown = false;
		
	g_mutex_lock(step_mutex);
	g_cond_wait(step_cond, step_mutex);
	g_mutex_unlock(step_mutex);
	GtkStatus("");
}

//****************************

void GRAPHICALUI::InputWait (char* buffer, unsigned len) {

	IF_DESTROYED_EXIT();
  
	inputDone = false;
	GtkStatus("Waiting for input");
	g_mutex_lock(wait_mutex);
	g_cond_wait(wait_cond, wait_mutex);
	g_mutex_unlock(wait_mutex);
	GtkStatus("");
	
	if(input) {
		if(strlen(input) >= len)
			input[len - 1] = '\0';
		sprintf(buffer, "%s", input);
	}
	else
		*buffer = '\0';

	//for the echo
	//
	if(*buffer)
		GRAPHICALUI::Output(MakeString("%s\n", buffer).c_str());

	free(input);
	input = (char *) 0;
}

//****************************

void GRAPHICALUI::Output (const char* msg) {
	IF_DESTROYED_EXIT();
	
	g_mutex_lock(visual_mutex);
	g_cond_wait(visual_cond, visual_mutex);
	if(!GtkDone) 
		GtkTextViewAddText(GTK_TEXT_VIEW(OutputText), msg);
	g_mutex_unlock(visual_mutex);
}

//****************************

void GRAPHICALUI::Trace (const char* msg) {
	IF_DESTROYED_EXIT();
  
	static char str[1024];
	sprintf(str, "%s.\n", msg);
	g_mutex_lock(visual_mutex);
	g_cond_wait(visual_cond, visual_mutex);
	if(!GtkDone)
		GtkTextViewAddText(GTK_TEXT_VIEW(TraceText), str);
	g_mutex_unlock(visual_mutex);
}


////////////////////////////////////////////////////
// Mutex wrapper function for using the GTK mutexes
//

unsigned GRAPHICALUI::CreateMutex (void)
	{ return (unsigned) g_mutex_new(); }

void GRAPHICALUI::DestroyMutex (unsigned id)
	{ g_mutex_free((GMutex *) id); }

void GRAPHICALUI::GrantMutex (unsigned id)
	{ g_mutex_lock((GMutex *) id); }

void GRAPHICALUI::ReleaseMutex (unsigned id)
	{ g_mutex_unlock((GMutex *) id); }


//********************************************************************

//////////////////////////////////////////////////////////////////////
// The functions that follow are part of the UI implementation. They
// are functions for our private use and therefore declared static.
//

//********************************************************************

////////////////////////////
// GUI callback functions
//

#ifdef SUPPLY_VISUALIZER_ACCESS_FUNCTIONS
///////////////////////////////////////////////////////////////////////
// Register and Memory access functions
//

static const char * RemoveSpaces (const char *str) {
	static char buf[1024];
	unsigned i, j;

	assert(str);

	for(i = 0, j = 0; str[i] != '\0'; ++i)
		if(!isspace(str[i]))
			buf[j++] = str[i];
	buf[j] = '\0';
	return buf;
}

//****************************

static void OnRegActivate (GtkEntry *entry, gpointer user_data) {
	unsigned num = (unsigned) user_data;
	int sign = 1;
	const char *text = RemoveSpaces(gtk_entry_get_text(entry));

	if(text[0] == '-') {
		sign = -1;
		++text;
	}

	for(unsigned i = 0; text[i] != '\0'; ++i)
		if(!isdigit(text[i])) {	//only allow decimal numbers
			GtkTextViewAddText(
				GTK_TEXT_VIEW(OutputText), 
				MakeString("%s\n", Configurator::RegAssignWarn).c_str()
			);
			GRAPHICALUI::RegisterRefresh(num);
			return;
		}

	int value = sign * atoi(text);

	SetJetReg(num, value);
	DataMirror::GetSingleton()->GrantRegAccess();
	DataMirror::GetSingleton()->SetRegContent(num, value);
	DataMirror::GetSingleton()->ReleaseRegAccess();

	GRAPHICALUI::RegisterRefresh(num);
}

//****************************

static void OnMemActivate (GtkEntry *entry, gpointer user_data) {
	bool error = false;
	unsigned pos = (unsigned) user_data;
	const char *text = RemoveSpaces(gtk_entry_get_text(entry));

	for(unsigned i = 0; text[i] != '\0'; ++i)
		if(!isdigit(text[i])) {	//only allow decimal numbers
			GtkTextViewAddText(
				GTK_TEXT_VIEW(OutputText), 
				MakeString("%s\n", Configurator::MemAssignWarnInvalid).c_str()
			);
			error = true;
			break;
		}

	int int_value = atoi(text);
	if(int_value >= 256) {
		GtkTextViewAddText(
			GTK_TEXT_VIEW(OutputText), 
			MakeString("%s\n", Configurator::MemAssignWarnOutOfRange).c_str()
			);
		error = true;
	}

	if(!error)
	{
		unsigned index = pos + CurrentPage * PAGE_SIZE;
		SetJetMem(index, (unsigned char) int_value);

		DataMirror::GetSingleton()->GrantMemAccess();
		DataMirror::GetSingleton()->SetMemContent(index, (unsigned char) int_value);
		DataMirror::GetSingleton()->ReleaseMemAccess();
	}
	
	GRAPHICALUI::MemoryRefresh(pos);
}
#endif

//****************************

#ifdef LARGE_MEMORY
static void OnPreviousClicked (GtkButton *button, gpointer user_data) {
	if(CurrentPage)
		GtkSetCurrentPage(--CurrentPage);
}

static void OnNextClicked (GtkButton *button, gpointer user_data) {
	if(CurrentPage != MEMORY_SIZE / PAGE_SIZE - 1)
		GtkSetCurrentPage(++CurrentPage);
}
#endif

/////////////////////////////////////////////////////////////////////////
// GTKDISPLAY namespace implementation functions

static void GTKDISPLAY::DisplayRegAction (GtkAction *action, GtkRadioAction *current) {
	gint value = gtk_radio_action_get_current_value (GTK_RADIO_ACTION (current));
	SetDisplay(REGS, value);
}

static void GTKDISPLAY::DisplayMemAction (GtkAction *action, GtkRadioAction *current) {
	gint value = gtk_radio_action_get_current_value (GTK_RADIO_ACTION (current));
	SetDisplay(MEM, value);
}

//****************************

static char* GTKDISPLAY::DisplayBin (int x, unsigned digits) {
	static char temp[BITS_PER_REGISTER + 1] = "";

	for(unsigned i = 0; i < digits; ++i) {
		int bit = x & (1 << i);
		temp[digits - 1 - i] = bit ? '1':'0';
	}
	temp[digits] = '\0';

	return temp;
}

static char* GTKDISPLAY::DisplayDec (int x, unsigned digits) {
	static char temp[32];
	sprintf(temp, "%d", x);
	return temp;
}

static char* GTKDISPLAY::DisplayHex (int x, unsigned digits) {
	// display_bin returns a static buffer which we can change.
	char* temp = DisplayBin(x, digits);
	unsigned NumOfBytes = digits >> 2;
	unsigned i;
	
	for(i = 0; i < NumOfBytes; ++i) {
		int res = 0;
		for(unsigned j = 0; j < 4; ++j)
			if(temp[4 * i + j] == '1')
				res += 1 << (3 - j);
		
		if(res <= 9)
			temp[i] = res + '0';
		else
			temp[i] = res - 10 + 'A';
	}
	temp[i] = '\0';
	return temp;
}

//****************************

static void GTKDISPLAY::SetDisplay (unsigned char whichDisplay, unsigned char whatDisplay) {
	// which : 0 -> Regs, 1 Mem
	// what  : 0 -> Bin, 1->Dec, 2->Hex
	switch(whichDisplay) {
		case REGS:		switch(whatDisplay) {
							case BIN:		displayRegs = DisplayBin;
											break;
							case DEC:		displayRegs = DisplayDec;
											break;
							case HEX:		displayRegs = DisplayHex;
											break;
							default:		assert(false);
						}
						RefreshAllRegisters();
						break;

		case MEM:		switch(whatDisplay)	{
							case BIN:		displayMem = DisplayBin;
											break;
							case HEX:		displayMem = DisplayHex;
											break;
							default:		assert(false);
						}
						RefreshAllMemory();
						break;

		default:		assert(false);
	}
}


//****************************

static void OnStepClicked (GtkButton *button, gpointer user_data) {
	g_mutex_lock(step_mutex);
	g_cond_signal(step_cond);
	g_mutex_unlock(step_mutex);
}

//****************************

static void OnInputActivate (GtkEntry *entry, gpointer user_data) {
	g_mutex_lock(wait_mutex);
	input = strdup((char *) gtk_entry_get_text(entry));
	gtk_entry_set_text(entry, "");
	inputDone = true;
	g_cond_signal(wait_cond);
	g_mutex_unlock(wait_mutex);
}

/////////////////////////////////////
// GTKMENU namespace implementation
//

static void GTKMENU::SetNotifyingState (GtkAction *action) {	
	const char *name = gtk_action_get_name (action);
	EVENTAPI::SetNotifyingEvent(
		name, 
		!EVENTAPI::IsNotifyingEvent(name)
	);
}


static void GTKMENU::SetSteppingState (GtkAction *action) {	
	const char *name = gtk_action_get_name (action);
	assert(!strcmp(name + strlen(name) - strlen("Step"), "Step"));	//must end with "Step"

	std::string s = name;
	s = s.substr(0, strlen(name) - strlen("Step"));	//cut the trailing "Step"
	EVENTAPI::SetSteppingEvent(
		s,
		!EVENTAPI::IsSteppingEvent(s)
	);
}

/////////////////////////////////////////////
// Functions for saving the execution TraceText
//

static void GTKMENU::Save (const char * name) {
	char temp[32];
	unsigned i;
	FILE *fp = fopen(name, "w");

	if(fp) {
		fputs("JET VISUALIZER : ", fp);
		fputs("\n\n\tOUTPUT\n", fp);
		fputs(GtkTextViewGetText(GTK_TEXT_VIEW(OutputText)), fp);
		fputs("\n\n\tPROGRAM TRACE\n", fp);
		fputs(GtkTextViewGetText(GTK_TEXT_VIEW(TraceText)), fp);

		fputs("\n\n\tREGISTERS\n", fp);
		for(i = 0; i < TOTAL_REGISTERS; ++i) {
			sprintf(temp, "R%d = ", i);
			fputs(temp, fp);
			fputs(gtk_entry_get_text(GTK_ENTRY(regs[i][1])), fp);
			fputs("\n", fp);
		}

		fputs("\n\n\t\tMEMORY\n", fp);
		for(i = 0; i < MEMORY_SIZE >> 2; ++i) {
			sprintf(temp, "%d\t", i << 2);
			fputs(temp, fp);
			for(unsigned j = 0; j < 4; ++j) {
				unsigned index = 4 * i + 3 - j;
				int content = 0;

				DataMirror::GetSingleton()->GrantMemAccess();
				if(DataMirror::GetSingleton()->IsMemOccupied(index))
					content = DataMirror::GetSingleton()->GetMemContent(index);
				DataMirror::GetSingleton()->ReleaseMemAccess();
	
				fputs(GTKDISPLAY::displayMem(content, 8), fp);
				fputs(" ", fp);
			}
			fputs("\n", fp);
		}
		fclose(fp);
	}
	else {
		GtkWidget *dialog;
		dialog = gtk_message_dialog_new (GTK_WINDOW(window),
                                  GTK_DIALOG_DESTROY_WITH_PARENT,
                                  GTK_MESSAGE_ERROR,
                                  GTK_BUTTONS_CLOSE,
                                  "Error creating file '%s': %s",
                                  name, g_strerror (errno));
		gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (dialog);
	}
}

//****************************

static void GTKMENU::StoreFilename (GtkWidget *widget, gpointer user_data) {
   GtkWidget *file_selector = GTK_WIDGET (user_data);
   SelectedFilename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_selector));
   
   Save(SelectedFilename);
}

//****************************

static void GTKMENU::CreateFileSelection (void) {
	GtkWidget *file_selector = gtk_file_selection_new ("Save as");

	g_signal_connect (
		GTK_FILE_SELECTION (file_selector)->ok_button,
		 "clicked",
		 G_CALLBACK (StoreFilename),
		 file_selector
	);

	g_signal_connect_swapped (
		GTK_FILE_SELECTION (file_selector)->ok_button,
		"clicked",
		G_CALLBACK (gtk_widget_destroy), 
		file_selector
	);

	g_signal_connect_swapped (
		GTK_FILE_SELECTION (file_selector)->cancel_button,
		"clicked",
		G_CALLBACK (gtk_widget_destroy),
		file_selector
	); 

	gtk_widget_show (file_selector);
}

////////////////////////////////////////////////
// Functions for handling the Menu bar actions
//

static void GTKMENU::File (GtkAction *action) {
	const char *name = gtk_action_get_name(action);

	if(!strcmp(name, "Save")) {
		if(SelectedFilename)
			Save(SelectedFilename);
		else
			CreateFileSelection();
	}
	else if(!strcmp(name, "SaveAs"))
		CreateFileSelection();
	else if(!strcmp(name, "Quit"))
		GtkDone = true;
	else
		assert(false);
}

//****************************

static void GTKMENU::ActivateEmail (GtkAboutDialog *about, const gchar *link, gpointer data) {
#ifndef WIN32
	char buf[1024];
	printf("Write your message and press CTRL+D to exit (noted by ^D) :\n");
	sprintf(buf, "mail %s", link);
	system(buf);
#else
	g_print("Unable to send mail to %s\n", link);
#endif
}

//****************************

static void GTKMENU::About (GtkAction *action, GtkWidget *window) {
	const gchar *author[] = {
		"HCI Laboratory, FORTH-ICS, 2005",
		"Concept, architecture and programming : Anthony Savidis",
		"E-mail: <as@ics.forth.gr>",
		"Co-programming and UI: John Lilis",
		NULL
	};

	const gchar *documentors[] = {
		"User guide: John Lilis",
		"Further contact: Anthony Savidis, FORTH-ICS",
		"<as@ics.forth.gr>",
		NULL
	};


	const gchar *license =
		"This program is a research outcome of HCI laboratory\n"
		"at FORTH-ICS. Currently it is provided for educational\n"
		"purposes only, at the Department of Computer Science,\n"
		"University of Crete. To gain a free deployment license\n"
		"please contact Anthony Savidis, as@ics.forth.gr";
	
	gtk_about_dialog_set_email_hook (ActivateEmail, NULL, NULL);
	gtk_show_about_dialog (GTK_WINDOW (window),
							"name", "Jet Visualizer",
							"version", "1.0",
							"copyright", "Anthony Savidis. HCI Lab, FORTH-ICS, 2005.",
							"license", license,
							"comments", "Execution visualizer for the Jet language.",
							"authors", author,
							"documenters", documentors,
							NULL);
}

//****************************

static void GTKMENU::Create (GtkWidget *window, GtkWidget *content) {
	static GtkActionEntry entries[] = {
		{ "FileMenu", NULL, "_File" },
		{ "VisualizedActive", NULL, "Visualized _Events" },
		{ "VisualizedStep", NULL, "Visualized _Step" },
		{ "Options", NULL, "_Options"},
		{ "Registers", NULL, "_Registers"},		
		{ "Memory", NULL, "_Memory"},
		{ "HelpMenu", NULL, "_Help" },
		//name, stock id label
		
		{ "Save", GTK_STOCK_SAVE, "_Save","<control>S",  
			"Save current file", G_CALLBACK (File) },
		{ "SaveAs", GTK_STOCK_SAVE,	"Save _As...", NULL,     
			"Save to a file", G_CALLBACK (File) },
		{ "Quit", GTK_STOCK_QUIT, "_Quit", "<control>Q",     
			"Quit",	G_CALLBACK (File) },
		{ "About", NULL, "_About", "<control>A",     
			"About", G_CALLBACK (About) },
		//name, stock id, label, accelerator, tooltip
	};

	static guint n_entries = G_N_ELEMENTS (entries);

#define TOGGLE_ENTRIES(name, x, func, value)	\
	static GtkToggleActionEntry name[] = {\
		{ "RFormat" x, NULL, "R-format Instructions", NULL, "R-Format", G_CALLBACK (func), value },\
		{ "IFormat" x, NULL, "I-format Instructions", NULL, "I-Format", G_CALLBACK (func), value },\
		{ "LFormat" x, NULL, "L-format Instructions", NULL, "L-Format", G_CALLBACK (func), value },\
		{ "Branch" x, NULL, "Branches", NULL, "Branches", G_CALLBACK (func), value },\
		{ "Jump" x, NULL, "Jumps", NULL, "Jumps", G_CALLBACK (func), value },\
		{ "RegAssign" x, NULL, "Assignment to Registers", NULL, "Assignment to Registers", G_CALLBACK (func), value },\
		{ "MemAssign" x, NULL, "Assignment to Memory", NULL, "Assignment to Memory", G_CALLBACK (func), value },\
	};

	TOGGLE_ENTRIES(toggle_entries_active, "", SetNotifyingState, TRUE)
	TOGGLE_ENTRIES(toggle_entries_step, "Step", SetSteppingState, FALSE)

	static guint n_toggle_entries_active = G_N_ELEMENTS (toggle_entries_active);
	static guint n_toggle_entries_step = G_N_ELEMENTS (toggle_entries_step);

	static GtkRadioActionEntry display_reg_entries [] = {
		{ "BinReg", NULL, 	"_Binary", "<control>B", "Bin", BIN },
		{ "DecReg", NULL, "_Decimal", "<control>D","Dec", DEC },
		{ "HexReg", NULL, "_Hexadecimal", "<control>H", "Hex", HEX },
		// name, stock, label, accelerator, tooltip, value
	};
	
	static GtkRadioActionEntry display_mem_entries [] = {
		{ "BinMem", NULL, 	"B_inary", "<control>I", "Bin", BIN },
		{ "DecMem", NULL, "D_ecimal", "<control>E","Dec", DEC },
		{ "HexMem", NULL, "He_xadecimal", "<control>X", "Hex", HEX },
		/* name, stock, label, accelerator, tooltip, value*/
	};

	static guint n_display_reg_entries = G_N_ELEMENTS (display_reg_entries);
	static guint n_display_mem_entries = G_N_ELEMENTS (display_mem_entries);

	static const char * ui_info = 
	"<ui>"
	"  <menubar name='MenuBar'>"
	"    <menu action='FileMenu'>"
	"      <menuitem action='Save'/>"
	"      <menuitem action='SaveAs'/>"
	"      <separator/>"
	"      <menuitem action='Quit'/>"
	"    </menu>"
	
	"	<menu name='Active' action='VisualizedActive'>"
	"      <menuitem action='RFormat'/>"
	"      <menuitem action='IFormat'/>"
	"      <menuitem action='LFormat'/>"
	"      <menuitem action='Branch'/>"
	"      <menuitem action='Jump'/>"
	"    </menu>"

	"	<menu name='Step' action='VisualizedStep'>"
	"      <menuitem action='RFormatStep'/>"
	"      <menuitem action='IFormatStep'/>"
	"      <menuitem action='LFormatStep'/>"
	"      <menuitem action='BranchStep'/>"
	"      <menuitem action='JumpStep'/>"
	"      <separator/>"
	"      <menuitem action='RegAssignStep'/>"
	"      <menuitem action='MemAssignStep'/>"
	"    </menu>"

	"    <menu action='Options'>"
	"		<menu action='Registers'>"
	"			<menuitem action='BinReg'/>"
	"			<menuitem action='DecReg'/>"
	"			<menuitem action='HexReg'/>"
	"		</menu>"
	"		<menu action='Memory'>"
	"			<menuitem action='BinMem'/>"
	"			<menuitem action='HexMem'/>"
	"		</menu>"
	"    </menu>"

	"    <menu action='HelpMenu'>"
	"      <menuitem action='About'/>"
	"    </menu>"
	"  </menubar>"
	"</ui>";

	GtkWidget *table;
	GtkWidget *bar;
	GtkActionGroup *action_group;
	GtkUIManager *merge;
	GError *error = NULL;

	table = gtk_table_new (1, 3, FALSE);
	gtk_container_add (GTK_CONTAINER (window), table);

	action_group = gtk_action_group_new ("AppWindowActions");
	gtk_action_group_add_actions (action_group, entries, n_entries, window);
	gtk_action_group_add_toggle_actions (action_group, toggle_entries_active, 
											n_toggle_entries_active, NULL);
	gtk_action_group_add_toggle_actions (action_group, toggle_entries_step, 
											n_toggle_entries_step, NULL);
	gtk_action_group_add_radio_actions (action_group, display_reg_entries, 
										n_display_reg_entries, DEC, 
										G_CALLBACK (GTKDISPLAY::DisplayRegAction), NULL);
	gtk_action_group_add_radio_actions (action_group, display_mem_entries, 
										n_display_mem_entries, BIN, 
										G_CALLBACK (GTKDISPLAY::DisplayMemAction), NULL);

	merge = gtk_ui_manager_new ();
	g_object_set_data_full (G_OBJECT (window), "ui-manager", merge, g_object_unref);
	gtk_ui_manager_insert_action_group (merge, action_group, 0);
	gtk_window_add_accel_group (GTK_WINDOW (window), 
		gtk_ui_manager_get_accel_group (merge));

	gtk_ui_manager_set_add_tearoffs (merge, TRUE);


	if (!gtk_ui_manager_add_ui_from_string (merge, ui_info, -1, &error)) {
		g_message ("Error building menus : %s", error->message);
		g_error_free (error);
		exit(1);
	}

	bar = gtk_ui_manager_get_widget (merge, "/MenuBar");
	gtk_widget_show (bar);
	gtk_table_attach (GTK_TABLE (table), bar, 0, 1, 0, 1,
		(GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 
		(GtkAttachOptions)(0), 0, 0);

	gtk_table_attach (GTK_TABLE (table), content, 0, 1, 2, 3,
		(GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 
		(GtkAttachOptions)(GTK_EXPAND | GTK_FILL), 0, 0);

	if (!GTK_WIDGET_VISIBLE (window))
		gtk_widget_show_all (window);
	else {    
		gtk_widget_destroy (window);
		window = NULL;
	}
}

//////////////////////////////////////////////
// GUI implementation helper & core functions.
//

static void GtkEnd (void) {
	GtkDone = true;
}

//****************************

static void GtkStatus (const char *str) {
	if(!GtkDone) {
		g_mutex_lock(visual_mutex);
		g_cond_wait(visual_cond, visual_mutex);
		gtk_statusbar_push(GTK_STATUSBAR(statusbar), 0, str);
		g_mutex_unlock(visual_mutex);
	}
}

//****************************

static void* GtkMain (void *unused) {
	//We can't get actual arguments as it is instantiated before main.
	//
	int argc = 0;
	char **argv = NULL;

    gtk_set_locale();
    gtk_init(&argc, &argv);
	
    window = GtkCreateWindow();
	gtk_widget_show (window);

	g_signal_connect ((gpointer) window, "destroy", G_CALLBACK(GtkEnd), NULL);

	g_mutex_lock(wait_mutex);
	g_cond_signal(wait_cond);	//signal the main thread to continue
	g_mutex_unlock(wait_mutex);

	//visual loop
	while(!GtkDone) {
		gdk_threads_enter();
		while(gtk_events_pending() && !GtkDone) {
			g_mutex_lock(visual_mutex);
			gtk_main_iteration();
			g_cond_signal(visual_cond);
			g_mutex_unlock(visual_mutex);
		}
		gdk_threads_leave();
	}

	window = (GtkWidget *) 0;
	
	g_cond_signal(visual_cond);
	g_cond_signal(step_cond);
	g_cond_signal(wait_cond);
	/*
	g_mutex_free(step_mutex);
	g_mutex_free(wait_mutex);
	g_mutex_free(visual_mutex);
	g_cond_free(step_cond);
	g_cond_free(wait_cond);
	g_cond_free(visual_cond);
	*/
	gtk_exit(0);
	exit(0);
    return (void *) 0;
}

//****************************

#ifdef LARGE_MEMORY
static void GtkSetCurrentPage(unsigned page) {
	char temp[50];
	unsigned i;
	for(i = 0; i < PAGE_SIZE >> 2; ++i) {
		sprintf(temp, "%d", page * PAGE_SIZE + (i << 2));
		gtk_label_set_text(GTK_LABEL(memory[i][0]), temp);
	}

	for(i = 0; i < PAGE_SIZE; ++i)
		GRAPHICALUI::MemoryRefresh(i);
}
#endif

//****************************

static const char *GtkTextViewGetText (GtkTextView *textview) {
	GtkTextBuffer *buf = gtk_text_view_get_buffer((GtkTextView *)textview);
	GtkTextIter start_iter, end_iter;
	
	gtk_text_buffer_get_start_iter(buf, &start_iter);
	gtk_text_buffer_get_end_iter(buf, &end_iter);

	return gtk_text_buffer_get_text(buf, &start_iter, &end_iter, FALSE);
}

//****************************

static void GtkTextViewAddText(GtkTextView *textview, const char *msg){
	GtkTextBuffer *buf = gtk_text_view_get_buffer(textview);
	GtkTextIter iter;

	gtk_text_buffer_get_end_iter(buf, &iter);
	gtk_text_view_scroll_to_iter(textview, &iter, 0, 1, 0, 0);
	gtk_text_buffer_insert(buf, &iter, msg, -1);
}

////////////////////////////////////////
// The actual window creating function.
// Part of it is auto-generate by glade.
//

static GtkWidget* GtkCreateWindow (void) {
	unsigned i, j;
	char temp[100];

    GtkWidget *window1;
	GtkWidget *table1;
    GtkWidget *vbox1;
    GtkWidget *hbox1;
    GtkWidget *vbox2;
    GtkWidget *label1;
    GtkWidget *scrolledwindow1;
    GtkWidget *vbox3;
    GtkWidget *label2;
    GtkWidget *scrolledwindow2;
    GtkWidget *hbox2;
    GtkWidget *vbox4;
    GtkWidget *label3;
    GtkWidget *scrolledwindow3;
    GtkWidget *vpaned1;
    GtkWidget *hpaned1;
    GtkWidget *vbox7;
    GtkWidget *label7;
    GtkWidget *fixed1;
    GtkWidget *button1;
    GtkWidget *vbox6;
    GtkWidget *label6;
    GtkWidget *scrolledwindow4;
	GtkWidget *viewport1;
	GtkWidget *viewport2;
	GtkWidget *reg_table;
	GtkWidget *mem_table;

#ifdef LARGE_MEMORY
	GtkWidget *button2;
	GtkWidget *button3;
	GtkWidget* vbox5;
#endif

    window1 = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window1), "JET VISUALIZER");
	gtk_window_set_default_size(GTK_WINDOW(window1), WINDOW_WIDTH, WINDOW_HEIGHT);

	table1 = gtk_table_new (2, 1, FALSE);
	gtk_widget_show (table1);

	GTKMENU::Create(window1, table1);
	
	statusbar = gtk_statusbar_new ();
	gtk_table_attach (GTK_TABLE (table1), statusbar, 0, 1, 3, 4,
					(GtkAttachOptions)(GTK_EXPAND | GTK_FILL),  
					(GtkAttachOptions)0, 0, 0);
	gtk_widget_show(statusbar);

    vbox1 = gtk_vpaned_new();
	//the memory/register area consists of the 4/9 of the window
	gtk_paned_set_position(GTK_PANED(vbox1), 4 * WINDOW_HEIGHT / 9);
    gtk_widget_show(vbox1);
    
	gtk_table_attach (GTK_TABLE (table1), vbox1, 0, 1, 0, 1,
					(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
					(GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);

	hbox1 = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox1);

#ifdef LARGE_MEMORY
	vbox5 = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox5);
    gtk_box_pack_start(GTK_BOX(hbox1), vbox5, FALSE, FALSE, 0);

	button2 = gtk_button_new_with_mnemonic("PREVIOUS");
    gtk_widget_show(button2);
	gtk_box_pack_start(GTK_BOX(vbox5), button2, TRUE, FALSE, 0);

	button3 = gtk_button_new_with_mnemonic("NEXT");
    gtk_widget_show(button3);
	gtk_box_pack_start(GTK_BOX(vbox5), button3, TRUE, FALSE, 0);
#endif

	gtk_paned_pack1(GTK_PANED(vbox1), hbox1, FALSE, FALSE);
   
    vbox2 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox2);
    gtk_box_pack_start(GTK_BOX(hbox1), vbox2, TRUE, TRUE, 0);

    label1 = gtk_label_new("MEMORY");
    gtk_widget_show(label1);
    gtk_box_pack_start(GTK_BOX(vbox2), label1, FALSE, FALSE, 0);

    scrolledwindow1 = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(scrolledwindow1);
    gtk_box_pack_start(GTK_BOX(vbox2), scrolledwindow1, TRUE, TRUE, 0);

    vbox3 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox3);
    gtk_box_pack_start(GTK_BOX(hbox1), vbox3, TRUE, TRUE, 0);

    label2 = gtk_label_new("REGISTERS");
    gtk_widget_show(label2);
    gtk_box_pack_start(GTK_BOX(vbox3), label2, FALSE, FALSE, 0);

    scrolledwindow2 = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(scrolledwindow2);
    gtk_box_pack_start(GTK_BOX(vbox3), scrolledwindow2, TRUE, TRUE, 0);

    hbox2 = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox2);
	gtk_paned_pack2(GTK_PANED(vbox1), hbox2, FALSE, FALSE);
   
    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);
    gtk_box_pack_start(GTK_BOX(hbox2), vbox4, TRUE, TRUE, 0);

    label3 = gtk_label_new("PROGRAM TRACE");
    gtk_widget_show(label3);
    gtk_box_pack_start(GTK_BOX(vbox4), label3, FALSE, FALSE, 0);

    scrolledwindow3 = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(scrolledwindow3);
    gtk_box_pack_start(GTK_BOX(vbox4), scrolledwindow3, TRUE, TRUE, 0);
    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolledwindow3),
                                        GTK_SHADOW_IN);

    TraceText = gtk_text_view_new();
	gtk_text_view_set_editable(GTK_TEXT_VIEW(TraceText), FALSE);
    gtk_widget_show(TraceText);
    gtk_container_add(GTK_CONTAINER(scrolledwindow3), TraceText);

    vpaned1 = gtk_vpaned_new();
    gtk_widget_show(vpaned1);
    gtk_box_pack_start(GTK_BOX(hbox2), vpaned1, TRUE, TRUE, 0);
    gtk_paned_set_position(GTK_PANED(vpaned1), 34);

    hpaned1 = gtk_hpaned_new();
    gtk_widget_show(hpaned1);
    gtk_paned_pack1(GTK_PANED(vpaned1), hpaned1, FALSE, FALSE);
    gtk_paned_set_position(GTK_PANED(hpaned1), 408);

    vbox7 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox7);
    gtk_paned_pack1(GTK_PANED(hpaned1), vbox7, FALSE, FALSE);

    label7 = gtk_label_new("      INPUT");
    gtk_widget_show(label7);
    gtk_box_pack_start(GTK_BOX(vbox7), label7, FALSE, FALSE, 0);

    InputEntry = gtk_entry_new();
    gtk_widget_show(InputEntry);
    gtk_box_pack_start(GTK_BOX(vbox7), InputEntry, FALSE, FALSE, 0);

    fixed1 = gtk_fixed_new();
    gtk_widget_show(fixed1);
    gtk_paned_pack2(GTK_PANED(hpaned1), fixed1, FALSE, FALSE);

    button1 = gtk_button_new_with_mnemonic("STEP");
    gtk_widget_show(button1);
    gtk_fixed_put(GTK_FIXED(fixed1), button1, 2, 12);
    gtk_widget_set_size_request(button1, 48, 23);

    vbox6 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox6);
    gtk_paned_pack2(GTK_PANED(vpaned1), vbox6, FALSE, FALSE);

    label6 = gtk_label_new("OUTPUT          ");
    gtk_widget_show(label6);
    gtk_box_pack_start(GTK_BOX(vbox6), label6, FALSE, FALSE, 0);

    scrolledwindow4 = gtk_scrolled_window_new(NULL, NULL);
    gtk_widget_show(scrolledwindow4);
    gtk_box_pack_start(GTK_BOX(vbox6), scrolledwindow4, TRUE, TRUE, 0);
    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolledwindow4),
                                        GTK_SHADOW_IN);

    OutputText = gtk_text_view_new();
	gtk_text_view_set_editable(GTK_TEXT_VIEW(OutputText), FALSE);
    gtk_widget_show(OutputText);
    gtk_container_add(GTK_CONTAINER(scrolledwindow4), OutputText);

	viewport2 = gtk_viewport_new (NULL, NULL);
	gtk_widget_show (viewport2);
	gtk_container_add (GTK_CONTAINER (scrolledwindow2), viewport2);

	reg_table = gtk_table_new (TOTAL_REGISTERS, 2, FALSE);
	gtk_widget_show (reg_table);
	gtk_container_add (GTK_CONTAINER (viewport2), reg_table);

	for(i = 0; i < TOTAL_REGISTERS; ++i)
	{
		sprintf(temp, "R%d", i);
		regs[i][0] = gtk_label_new (temp);
		gtk_widget_show (regs[i][0]);
		gtk_table_attach (GTK_TABLE (reg_table), regs[i][0], 0, 1, i, i+1,
	    				 (GtkAttachOptions) (GTK_FILL),
						 (GtkAttachOptions) (0), 0, 0);
		gtk_widget_set_size_request(regs[i][0], 60, -1);

		regs[i][1] = gtk_entry_new ();

#ifndef SUPPLY_VISUALIZER_ACCESS_FUNCTIONS
		gtk_editable_set_editable(GTK_EDITABLE(regs[i][1]), FALSE);
#endif
		gtk_entry_set_text(GTK_ENTRY(regs[i][1]), "0");

		gtk_entry_set_alignment(GTK_ENTRY(regs[i][1]), 1);
		gtk_widget_show (regs[i][1]);
		gtk_table_attach (GTK_TABLE (reg_table), regs[i][1], 1, 2, i, i+1,
					   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
					   (GtkAttachOptions) (0), 0, 0);
	}
	
	//no access to register 0 as it is hardwired
	gtk_editable_set_editable(GTK_EDITABLE(regs[0][1]), FALSE);


	viewport1 = gtk_viewport_new (NULL, NULL);
	gtk_widget_show (viewport1);
	gtk_container_add (GTK_CONTAINER (scrolledwindow1), viewport1);

	unsigned mem_limit = PAGE_SIZE >> 2;

	mem_table = gtk_table_new (mem_limit, 5, FALSE);
	gtk_widget_show (mem_table);
	gtk_container_add (GTK_CONTAINER (viewport1), mem_table);

	for(i = 0; i < mem_limit; ++i)
	{
		for(j = 1; j < 5; ++j)
		{
			memory[i][j] = gtk_entry_new ();

#ifndef SUPPLY_VISUALIZER_ACCESS_FUNCTIONS
			gtk_editable_set_editable(GTK_EDITABLE(memory[i][j]), FALSE);
#endif
			gtk_entry_set_text(GTK_ENTRY(memory[i][j]), "00000000");

			gtk_entry_set_alignment(GTK_ENTRY(memory[i][j]), 0.5);
			gtk_widget_show (memory[i][j]);
			gtk_table_attach (GTK_TABLE (mem_table), memory[i][j], j, j+1, i, i+1,
						   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
						   (GtkAttachOptions) (0), 0, 0);
			gtk_widget_set_size_request(memory[i][j], 40, -1);
		}
	}

	for(i = 0; i < mem_limit; ++i)
	{
		sprintf(temp, "%d", i << 2);
		memory[i][0] = gtk_label_new (temp);
		gtk_widget_show (memory[i][0]);
		gtk_table_attach (GTK_TABLE (mem_table), memory[i][0], 0, 1, i, i+1,
	    				 (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
						 (GtkAttachOptions) (0), 0, 0);
		gtk_widget_set_size_request(memory[i][0], 10, -1);
	}


	g_signal_connect ((gpointer) button1, "clicked",
					G_CALLBACK (OnStepClicked), NULL);
	g_signal_connect ((gpointer) InputEntry, "activate",
					G_CALLBACK (OnInputActivate), NULL);

#ifdef SUPPLY_VISUALIZER_ACCESS_FUNCTIONS
	//no access to register 0 as it is hardwires to zero
	for(i = 1; i < TOTAL_REGISTERS; ++i)
		g_signal_connect ((gpointer) regs[i][1], "activate",
						G_CALLBACK (OnRegActivate), (void *) i);

	for(i = 0; i < mem_limit; ++i)
		g_signal_connect ((gpointer) memory[i >> 2][4 - (i & 3)], "activate",
						G_CALLBACK (OnMemActivate), (void *) i);
#endif
#ifdef LARGE_MEMORY
	g_signal_connect ((gpointer) button2, "clicked",
					G_CALLBACK (OnPreviousClicked), NULL);
	g_signal_connect ((gpointer) button3, "clicked",
					G_CALLBACK (OnNextClicked), NULL);
#endif

    return window1;
}

//****************************

void RefreshAllRegisters (void) {
	for(unsigned i = 0; i < TOTAL_REGISTERS; ++i)
		GRAPHICALUI::RegisterRefresh(i);
}


void RefreshAllMemory (void) {
	unsigned limit = (CurrentPage + 1) * PAGE_SIZE;
	for(unsigned i = CurrentPage * PAGE_SIZE; i < limit; ++i)
		GRAPHICALUI::MemoryRefresh(i);
}

#endif	// HAS_VISUALIZER && GRAPHICAL_VISUALIZER
