// GraphicalVisualiser.cpp
// A graphical visualiser for the JET language.
// A. Savidis, Y.Lilis.
// November, 2005.
// Project: JET Suite (Execution visualizer, Savidis, 2005).
//

#if	defined(HAS_VISUALIZER) && defined(GRAPHICAL_VISUALIZER)

#include "GraphicalVisualiser.h"
#include "GraphicalUILinkage.h"
#include "DataMirror.h"

#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

VISUALISER_LINKAGE(GraphicalVisualiser)

////////////////////////////////////////////////////////////////

const char* GraphicalVisualiser::Input (const char* buff, unsigned len, FILE* fp) {

	// We ingore fp.
	//
	GRAPHICALUI::InputWait(const_cast<char*>(buff), len);
	return buff;
}

// Only program output.
//
int GraphicalVisualiser::Output (const char* format, ...) {

	va_list args;
	va_start(args, format);

	int result;
	static char buffer[1024];
	result = vsprintf(buffer, format, args);

	GRAPHICALUI::Output(buffer);

	va_end(args);
	return result;
}

////////////////////////////////////////////////////////////////

void GraphicalVisualiser::PreInitialiser (void) {
	GRAPHICALUI::Create();
	DataMirror::SingletonCreate();
}

void GraphicalVisualiser::PostCleaner (void) {
	GRAPHICALUI::Destroy();
	DataMirror::SingletonDestroy();
}

void GraphicalVisualiser::CleanUp (void) {

	assert(!cleanedUp && visualiser);

	cleanedUp = true;
	delete ((GraphicalVisualiser*) visualiser);	// Derived class specific.
	visualiser = (VisualiserSuper*) 0;

	PostCleaner();
	exit(0);								// Clean-up exits immediately!
}

////////////////////////////////////////////////////////////////
// On program start, we explicitly wait upon start / end / stop.
//
void GraphicalVisualiser::OnProgramStart (void) {
	Output("Servant starts running the program.\n");

	GRAPHICALUI::Started();
	GetVisualiser()->StepWait();
}

void GraphicalVisualiser::OnProgramEnd (void) {
	Output("Servant finished the program normally");

	GRAPHICALUI::Finished();

	GetVisualiser()->StepWait();
	CleanUp();
}

void GraphicalVisualiser::OnProgramError (const char* s) {
	GRAPHICALUI::Trace(
		MakeString(
			"Error: %s\nServant has interrupted program execution", s
		).c_str()
	);

	GRAPHICALUI::Finished();

	GetVisualiser()->StepWait();
	CleanUp();
}

////////////////////////////////////////////////////////////////

GraphicalVisualiser::GraphicalVisualiser (void) {

	PreInitialiser();

	eventMutex = GRAPHICALUI::CreateMutex();

	ExecutionEvents::SetInputFunction(Input);
	ExecutionEvents::SetOutputFunction(Output);
	ExecutionEvents::SetCleanUpFunction(CleanUp);

	ExecutionEvents::SetOnProgramStart(OnProgramStart);
	ExecutionEvents::SetOnProgramEnd(OnProgramEnd);
	ExecutionEvents::SetOnProgramError(OnProgramError);

	GRAPHICALUI::Trace(
		MakeString(
			"Servant starts running the program.\n"
			"Graphical execution tracer activated"
		).c_str()
	);
}

GraphicalVisualiser::~GraphicalVisualiser() {
	GRAPHICALUI::DestroyMutex(eventMutex );
}

////////////////////////////////////////////////////////////////

bool GraphicalVisualiser::IsNotifyingEvent (const std::string& name) {
	GRAPHICALUI::GrantMutex(eventMutex);
	bool result = VisualiserSuper::IsNotifyingEvent(name);
	GRAPHICALUI::ReleaseMutex(eventMutex);
	return result;
}

bool EVENTAPI::IsNotifyingEvent (const std::string& name)
	{ return VisualiserSuper::GetVisualiser()->IsNotifyingEvent(name); }

//****************************

void GraphicalVisualiser::SetNotifyingEvent (const std::string& name, bool enabled) {
	GRAPHICALUI::GrantMutex(eventMutex);
	VisualiserSuper::SetNotifyingEvent(name, enabled);
	GRAPHICALUI::ReleaseMutex(eventMutex);
}

void EVENTAPI::SetNotifyingEvent (const std::string& name, bool enabled)
	{ VisualiserSuper::GetVisualiser()->SetNotifyingEvent(name, enabled); }

//****************************

bool GraphicalVisualiser::IsSteppingEvent (const std::string& name) {
	GRAPHICALUI::GrantMutex(eventMutex);
	bool result = VisualiserSuper::IsSteppingEvent(name);
	GRAPHICALUI::ReleaseMutex(eventMutex);
	return result;
}

bool EVENTAPI::IsSteppingEvent (const std::string& name)
	{ return VisualiserSuper::GetVisualiser()->IsSteppingEvent(name); }

//****************************

void GraphicalVisualiser::SetSteppingEvent (const std::string& name, bool enabled) {
	GRAPHICALUI::GrantMutex(eventMutex);
	VisualiserSuper::SetSteppingEvent(name, enabled);
	GRAPHICALUI::ReleaseMutex(eventMutex);
}

void EVENTAPI::SetSteppingEvent (const std::string& name, bool enabled)
	{ VisualiserSuper::GetVisualiser()->SetSteppingEvent(name, enabled); }

////////////////////////////////////////////////////////////////

void GraphicalVisualiser::Message (const char* str) const
	{ GRAPHICALUI::Trace(str); }

//****************************

void GraphicalVisualiser::StepWait (void) const 
	{ GRAPHICALUI::StepWait(); }

//****************************

void GraphicalVisualiser::MemCellRefresh (unsigned MemCell, int value) {
	assert(visualiser);

	DataMirror::GetSingleton()->GrantMemAccess();
	DataMirror::GetSingleton()->SetMemContent(MemCell, value);
	DataMirror::GetSingleton()->ReleaseMemAccess();

	Message(
		MakeString(
			"Memory assignment : MEM[%u] = %d", 
			MemCell,
			value
		).c_str()
	);
	// Now ask to refresh content.
	//
	GRAPHICALUI::MemoryRefresh(MemCell);
}

//****************************

void GraphicalVisualiser::RegRefresh (unsigned reg, int value) {
	assert(visualiser);

	DataMirror::GetSingleton()->GrantRegAccess();
	DataMirror::GetSingleton()->SetRegContent(reg, value);
	DataMirror::GetSingleton()->ReleaseRegAccess();

	Message(
		MakeString(
			"Register assignment : REG[%u] = %d",
			reg,
			value
		).c_str()
	);

	// Again, ask to refresh content.
	//
	GRAPHICALUI::RegisterRefresh(reg);
}

////////////////////////////////////////////////////////////////

#endif	// HAS_VISUALIZER && GTK_VISUALIZER
